@@ -33,14 +33,68 @@ http://www.ncdc.noaa.gov/oa/pod-guide/ncdc/docs/klm/html/c8/sec83142-1.htm
 
 import datetime
 import logging
+try:
+    from enum import IntFlag
+except ImportError:
+    # python version < 3.6, use a simple object without nice representation
+    IntFlag = object
 
 import numpy as np
 
 from pygac.correct_tsm_issue import TSM_AFFECTED_INTERVALS_KLM, get_tsm_idx
-from pygac.reader import Reader
+from pygac.reader import Reader, ReaderError
+from pygac.utils import file_opener
 
 LOG = logging.getLogger(__name__)
 
+
+class KLM_QualityIndicator(IntFlag):
+    """Quality Indicators.
+
+    Source:
+        KLM guide
+        Table 8.3.1.3.3.1-1. Format of packed LAC/HRPT Data Sets (Version 2, pre-April 28, 2005).
+        Table 8.3.1.3.3.2-1. Format of LAC/HRPT Data Record for NOAA-N (Version 5, post-November 14,
+                             2006, all spacecraft).
+        Table 8.3.1.4.3.1-1. Format of packed GAC Data Record for NOAA KLM (Version 2, pre-April 28, 2005).
+        Table 8.3.1.4.3.2-1. Format of GAC Data Record for NOAA-N (Version 4, post-January 25, 2006,
+                             all spacecraft).
+
+    Note:
+        Table 8.3.1.3.3.1-1. and Table 8.3.1.4.3.1-1. define bit: 21 as
+        "frame sync word not valid"
+        Table 8.3.1.3.3.2-1. and Table 8.3.1.4.3.2-1. define bit: 21 as
+        "flywheeling detected during this frame"
+    """
+    FATAL_FLAG = 2**0  # Data should not be used for product generation
+    TIME_ERROR = 2**1  # Time sequence error detected within this scan
+    DATA_GAP = 2**2  # Data gap precedes this scan
+    CALIBRATION = 2**3  # Insufficient data for calibration
+    NO_EARTH_LOCATION = 2**4  # Earth location data not available
+    CLOCK_UPDATE = 2**5  # First good time following a clock update (nominally 0)
+    INSTRUMENT_CHANGE = 2**6  # Instrument status changed with this scan
+    BIT_SYNC_STATUS = 2**7  # Sync lock dropped during this frame
+    SYNC_ERROR = 2**8  # Frame sync word error greater than zero
+    FRAME_SYNC_LOCK = 2**9  # Frame sync previously dropped lock
+    SYNC_INVALID = 2**10  # Frame sync word not valid
+    FLYWHEELING = 2**10  # Flywheeling detected during this frame
+    BIT_SLIPPAGE = 2**11  # Bit slippage detected during this frame
+    # Note: 2**12 to 2**22 are not defined for KLMs
+    TIP_PARITY = 2**23  # TIP parity error detected
+    # Reflected Sunlight (RS) detected (solar blackbody contamination)
+    CH_3B_RS = 2**24
+    CH_3_CONTAMINATION = 2**24  # POD compatible alias
+    CH_3B_RS_ANOMALY = 2**25
+    CH_4_RS = 2**26
+    CH_4_CONTAMINATION = 2**26  # POD compatible alias
+    CH_4_RS_ANOMALY = 2**27
+    CH_5_RS = 2**28
+    CH_5_CONTAMINATION = 2**28  # POD compatible alias
+    CH_5_RS_ANOMALY = 2**29
+    DATA_JITTER = 2**30  # Resync occurred on this frame
+    PSEUDO_NOISE = 2**31  # Pseudo noise occurred on this frame
+
+
 # GAC header object
 
 header = np.dtype([("data_set_creation_site_id", "S3"),
@@ -574,12 +628,15 @@ class KLMReader(Reader):
 
     tsm_affected_intervals = TSM_AFFECTED_INTERVALS_KLM
 
-    def read(self, filename):
+    QFlag = KLM_QualityIndicator
+    _quality_indicators_key = "quality_indicator_bit_field"
+
+    def read(self, filename, fileobj=None):
         """Read the data.
 
         Args:
-            filename: string
-                The filename to read from.
+            filename: Path to GAC/LAC file
+            fileobj: An open file object to read from. (optional)
 
         Returns:
             header: numpy record array
@@ -588,23 +645,19 @@ class KLMReader(Reader):
                 The scanlines
 
         """
-        with self._open(filename) as fd_:
-            # Note that np.fromfile does not work with gzip.GzipFile
-            # objects (numpy version 1.16.4), because it restricts the
-            # file objects to (io.FileIO, io.BufferedReader, io.BufferedWriter)
-            # see: numpy.compat.py3k.isfileobj
-            self.ars_head, = np.frombuffer(
-                fd_.read(ars_header.itemsize),
-                dtype=ars_header, count=1)
-            if not self.ars_head['data_format'].startswith(b'NOAA Level 1b'):
-                fd_.seek(0)
-                self.ars_head = None
-                ars_offset = 0
-            else:
+        # Note that np.fromfile does not work with gzip.GzipFile
+        # objects (numpy version 1.16.4), because it restricts the
+        # file objects to (io.FileIO, io.BufferedReader, io.BufferedWriter)
+        # see: numpy.compat.py3k.isfileobj
+        self.filename = filename
+        LOG.info('Reading %s', self.filename)
+        with file_opener(fileobj or filename) as fd_:
+            self.ars_head, self.head = self.read_header(
+                filename, fileobj=fd_)
+            if self.ars_head:
                 ars_offset = ars_header.itemsize
-            self.head, = np.frombuffer(
-                fd_.read(header.itemsize),
-                dtype=header, count=1)
+            else:
+                ars_offset = 0
             self.header_version = self.head[
                 "noaa_level_1b_format_version_number"]
             if self.header_version >= 5:
@@ -617,16 +670,59 @@ class KLMReader(Reader):
                     dtype=analog_telemetry_v2, count=1)
             # LAC: 1, GAC: 2, ...
             self.data_type = self.head['data_type_code']
+            # read until end of file
             fd_.seek(self.offset + ars_offset, 0)
-            self.scans = np.frombuffer(
-                fd_.read(),
-                dtype=self.scanline_type,
-                count=self.head["count_of_data_records"])
-
+            buffer = fd_.read()
+            count = self.head["count_of_data_records"]
+            self._read_scanlines(buffer, count)
         self.correct_scan_line_numbers()
         self.spacecraft_id = self.head["noaa_spacecraft_identification_code"]
         self.spacecraft_name = self.spacecraft_names[self.spacecraft_id]
 
+    @classmethod
+    def read_header(cls, filename, fileobj=None):
+        """Read the file header.
+
+        Args:
+            filename (str): Path to GAC/LAC file
+            fileobj: An open file object to read from. (optional)
+
+        Returns:
+            archive_header (struct): archive header
+            header (struct): file header
+        """
+        with file_opener(fileobj or filename) as fd_:
+            # read ars_header if present
+            _ars_head, = np.frombuffer(
+                fd_.read(ars_header.itemsize),
+                dtype=ars_header, count=1)
+            if _ars_head['data_format'].startswith(b'NOAA Level 1b'):
+                ars_head = _ars_head.copy()
+            else:
+                fd_.seek(0)
+                ars_head = None
+            # need to copy frombuffer to have write access on head
+            head, = np.frombuffer(
+                fd_.read(header.itemsize),
+                dtype=header, count=1).copy()
+        head = cls._correct_data_set_name(head, filename)
+        cls._validate_header(head)
+        return ars_head, head
+
+    @classmethod
+    def _validate_header(cls, header):
+        """Check if the header belongs to this reader"""
+        # call super to enter the Method Resolution Order (MRO)
+        super(KLMReader, cls)._validate_header(header)
+        LOG.debug("validate header")
+        data_set_name = header['data_set_name'].decode()
+        # split header into parts
+        creation_site, transfer_mode, platform_id = (
+            data_set_name.split('.')[:3])
+        allowed_ids = ['NK', 'NL', 'NM', 'NN', 'NP', 'M1', 'M2', 'M3']
+        if platform_id not in allowed_ids:
+            raise ReaderError('Improper platform id "%s"!' % platform_id)
+
     def get_telemetry(self):
         """Get the telemetry.
 
@@ -695,32 +791,6 @@ class KLMReader(Reader):
         """
         return self.scans["scan_line_bit_field"][:] & 3
 
-    def _get_corrupt_mask(self):
-        """Get mask for corrupt scanlines."""
-        mask = ((self.scans["quality_indicator_bit_field"] >> 31) |
-                ((self.scans["quality_indicator_bit_field"] << 3) >> 31) |
-                ((self.scans["quality_indicator_bit_field"] << 4) >> 31))
-        return mask.astype(bool)
-
-    def get_qual_flags(self):
-        """Read quality flags."""
-        number_of_scans = self.scans["telemetry"].shape[0]
-        qual_flags = np.zeros((int(number_of_scans), 7))
-        qual_flags[:, 0] = self.scans["scan_line_number"]
-        qual_flags[:, 1] = (self.scans["quality_indicator_bit_field"] >> 31)
-        qual_flags[:, 2] = (
-            (self.scans["quality_indicator_bit_field"] << 3) >> 31)
-        qual_flags[:, 3] = (
-            (self.scans["quality_indicator_bit_field"] << 4) >> 31)
-        qual_flags[:, 4] = (
-            (self.scans["quality_indicator_bit_field"] << 24) >> 30)
-        qual_flags[:, 5] = (
-            (self.scans["quality_indicator_bit_field"] << 26) >> 30)
-        qual_flags[:, 6] = (
-            (self.scans["quality_indicator_bit_field"] << 28) >> 30)
-
-        return qual_flags
-
     def postproc(self, channels):
         """Apply KLM specific postprocessing.
 
@@ -747,28 +817,7 @@ class KLMReader(Reader):
 
 def main_klm(reader_cls, filename, start_line, end_line):
     """Generate a l1c file."""
-    from pygac import gac_io
     tic = datetime.datetime.now()
-    reader = reader_cls()
-    reader.read(filename)
-    reader.get_lonlat()
-    channels = reader.get_calibrated_channels()
-    sat_azi, sat_zen, sun_azi, sun_zen, rel_azi = reader.get_angles()
-    qual_flags = reader.get_qual_flags()
-    if (np.all(reader.mask)):
-        print("ERROR: All data is masked out. Stop processing")
-        raise ValueError("All data is masked out.")
-
-    gac_io.save_gac(reader.spacecraft_name,
-                    reader.utcs,
-                    reader.lats, reader.lons,
-                    channels[:, :, 0], channels[:, :, 1],
-                    channels[:, :, 2],
-                    channels[:, :, 3],
-                    channels[:, :, 4],
-                    channels[:, :, 5],
-                    sun_zen, sat_zen, sun_azi, sat_azi, rel_azi,
-                    qual_flags, start_line, end_line,
-                    reader.filename,
-                    reader.meta_data)
+    reader = reader_cls.fromfile(filename)
+    reader.save(int(start_line), int(end_line))
     LOG.info("pygac took: %s", str(datetime.datetime.now() - tic))
