@@ -36,14 +36,60 @@ http://www.ncdc.noaa.gov/oa/pod-guide/ncdc/docs/podug/html/c3/sec3-1.htm
 
 import datetime
 import logging
+try:
+    from enum import IntFlag
+except ImportError:
+    # python version < 3.6, use a simple object without nice representation
+    IntFlag = object
 
 import numpy as np
 
 from pygac.correct_tsm_issue import TSM_AFFECTED_INTERVALS_POD, get_tsm_idx
-from pygac.reader import Reader
+from pygac.reader import Reader, ReaderError
+from pygac.utils import file_opener
 
 LOG = logging.getLogger(__name__)
 
+
+class POD_QualityIndicator(IntFlag):
+    """Quality Indicators.
+
+    Source:
+        POD guide Table 3.1.2.1-2. Format of quality indicators.
+    """
+    # POD guide Table 3.1.2.1-2. Format of quality indicators.
+    FATAL_FLAG = 2**0  # Data should not be used for product generation
+    TIME_ERROR = 2**1  # A time sequence error was detected while Processing
+    # this frame
+    DATA_GAP = 2**2  # A gap precedes this frame
+    DATA_JITTER = 2**3  # Resync occurred on this frame
+    CALIBRATION = 2**4  # Insufficient data for calibration
+    NO_EARTH_LOCATION = 2**5  # Earth location data not available
+    ASCEND_DESCEND = 2**6  # AVHRR Earth location indication of Ascending (=0)
+    # or descending (=1) data
+    PSEUDO_NOISE = 2**7  # Pseudo Noise (P/N) occurred (=1) on the frame,
+    # data not used for calibration computations
+    BIT_SYNC_STATUS = 2**8  # Drop lock during frame
+    SYNC_ERROR = 2**9  # Frame Sync word error greater than zero
+    FRAME_SYNC_LOCK = 2**10  # Frame Sync previously dropped lock
+    FLYWHEELING = 2**11  # Flywheeling detected during this frame
+    BIT_SLIPPAGE = 2**12  # Bit slippage detected during this frame
+    # Solar blackbody contamination indicator
+    # 0 = no correction
+    # 1 = solar contamination corrected
+    CH_3_CONTAMINATION = 2**13  # Channel 3 solar blackbody contamination
+    CH_4_CONTAMINATION = 2**14  # Channel 4 solar blackbody contamination
+    CH_5_CONTAMINATION = 2**15  # Channel 5 solar blackbody contamination
+    # TIP Parity
+    TIP_PARITY_1 = 2**16  # In first minor frame
+    TIP_PARITY_2 = 2**17  # In second minor frame
+    TIP_PARITY_3 = 2**18  # In third minor frame
+    TIP_PARITY_4 = 2**19  # In fourth minor frame
+    TIP_PARITY_5 = 2**20  # In fifth minor frame
+    # Note: 2**21 to 2**23, and 2**30 to 2**31 are spare bits. 2**24 to 2**29 define
+    #       "SYNC ERRORS - Number of bit errors in frame sync" (6 bit integer?)
+
+
 # common header
 header0 = np.dtype([("noaa_spacecraft_identification_code", ">u1"),
                     ("data_type_code", ">u1"),
@@ -186,6 +232,9 @@ class PODReader(Reader):
 
     tsm_affected_intervals = TSM_AFFECTED_INTERVALS_POD
 
+    QFlag = POD_QualityIndicator
+    _quality_indicators_key = "quality_indicators"
+
     def correct_scan_line_numbers(self):
         """Correct the scan line numbers."""
         # Perform common corrections first.
@@ -203,12 +252,12 @@ class PODReader(Reader):
 
         self.scans = self.scans[self.scans["scan_line_number"] != 0]
 
-    def read(self, filename):
+    def read(self, filename, fileobj=None):
         """Read the data.
 
         Args:
-            filename: string
-                The filename to read from.
+            filename (str): Path to GAC/LAC file
+            fileobj: An open file object to read from. (optional)
 
         Returns:
             header: numpy record array
@@ -217,45 +266,24 @@ class PODReader(Reader):
                 The scanlines
 
         """
+        self.filename = filename
+        LOG.info('Reading %s', self.filename)
         # choose the right header depending on the date
-        with self._open(filename) as fd_:
-            # read archive header
-            self.tbm_head, = np.frombuffer(
-                fd_.read(tbm_header.itemsize),
-                dtype=tbm_header, count=1)
-            if ((not self.tbm_head['data_set_name'].startswith(self.creation_site + b'.')) and
-                    (self.tbm_head['data_set_name'] != b'\x00' * 42 + b'  ')):
-                fd_.seek(0)
-                self.tbm_head = None
-                tbm_offset = 0
-            else:
+        with file_opener(fileobj or filename) as fd_:
+            self.tbm_head, self.head = self.read_header(
+                filename, fileobj=fd_)
+            if self.tbm_head:
                 tbm_offset = tbm_header.itemsize
-
-            head, = np.frombuffer(
-                fd_.read(header0.itemsize),
-                dtype=header0, count=1)
-            year, jday, _ = self.decode_timestamps(head["start_time"])
-
-            start_date = (datetime.date(year, 1, 1) +
-                          datetime.timedelta(days=int(jday) - 1))
-
-            if start_date < datetime.date(1992, 9, 8):
-                header = header1
-            elif start_date <= datetime.date(1994, 11, 15):
-                header = header2
             else:
-                header = header3
-
-            fd_.seek(tbm_offset, 0)
-            self.head, = np.frombuffer(
-                fd_.read(header.itemsize),
-                dtype=header, count=1)
+                tbm_offset = 0
+            # read scan lines until end of file
             fd_.seek(self.offset + tbm_offset, 0)
-            self.scans = np.frombuffer(
-                fd_.read(),
-                dtype=self.scanline_type,
-                count=self.head["number_of_scans"])
-
+            buffer = fd_.read()
+            count = self.head["number_of_scans"]
+            self._read_scanlines(buffer, count)
+        year, jday, _ = self.decode_timestamps(self.head["start_time"])
+        start_date = (datetime.date(year, 1, 1) +
+                      datetime.timedelta(days=int(jday) - 1))
         self.correct_scan_line_numbers()
         self.spacecraft_id = self.head["noaa_spacecraft_identification_code"]
         if self.spacecraft_id == 1 and start_date < datetime.date(1982, 1, 1):
@@ -263,9 +291,76 @@ class PODReader(Reader):
         self.spacecraft_name = self.spacecraft_names[self.spacecraft_id]
         LOG.info(
             "Reading %s data", self.spacecrafts_orbital[self.spacecraft_id])
-
         return self.head, self.scans
 
+    @classmethod
+    def read_header(cls, filename, fileobj=None):
+        """Read the file header.
+
+        Args:
+            filename (str): Path to GAC/LAC file
+            fileobj: An open file object to read from. (optional)
+
+        Returns:
+            archive_header (struct): archive header
+            header (struct): file header
+        """
+        # choose the right header depending on the date
+        with file_opener(fileobj or filename) as fd_:
+            # read tbm_header if present
+            _tbm_head, = np.frombuffer(
+                fd_.read(tbm_header.itemsize),
+                dtype=tbm_header, count=1)
+            try:
+                data_set_name = _tbm_head['data_set_name'].decode()
+            except UnicodeDecodeError:
+                data_set_name = '---'
+            allowed_empty = (42*b'\x00' + b'  ')
+            if (cls.data_set_pattern.match(data_set_name)
+                    or (_tbm_head['data_set_name'] == allowed_empty)):
+                tbm_head = _tbm_head.copy()
+                tbm_offset = tbm_header.itemsize
+            else:
+                fd_.seek(0)
+                tbm_head = None
+                tbm_offset = 0
+            # read header
+            head0, = np.frombuffer(
+                fd_.read(header0.itemsize),
+                dtype=header0, count=1)
+            year, jday, _ = cls.decode_timestamps(head0["start_time"])
+            start_date = (datetime.date(year, 1, 1) +
+                          datetime.timedelta(days=int(jday) - 1))
+            if start_date < datetime.date(1992, 9, 8):
+                header = header1
+            elif start_date <= datetime.date(1994, 11, 15):
+                header = header2
+            else:
+                header = header3
+            fd_.seek(tbm_offset, 0)
+            # need to copy frombuffer to have write access on head
+            head, = np.frombuffer(
+                fd_.read(header.itemsize),
+                dtype=header, count=1).copy()
+        head = cls._correct_data_set_name(head, filename)
+        cls._validate_header(head)
+        return tbm_head, head
+
+    @classmethod
+    def _validate_header(cls, header):
+        """Check if the header belongs to this reader"""
+        # call super to enter the Method Resolution Order (MRO)
+        super(PODReader, cls)._validate_header(header)
+        LOG.debug("validate header")
+        data_set_name = header['data_set_name'].decode()
+        # split header into parts
+        creation_site, transfer_mode, platform_id = (
+            data_set_name.split('.')[:3])
+        allowed_ids = ['TN', 'NA', 'NB', 'NC', 'ND', 'NE', 'NF', 'NG',
+                       'NH', 'NI', 'NJ']
+        if platform_id not in allowed_ids:
+            raise ReaderError('Improper platform id "%s"!' % platform_id)
+
     def get_header_timestamp(self):
         """Get the timestamp from the header.
 
@@ -432,27 +527,6 @@ class PODReader(Reader):
 
         return prt_counts, ict_counts, space_counts
 
-    def _get_corrupt_mask(self):
-        """Get mask for corrupt scanlines."""
-        mask = ((self.scans["quality_indicators"] >> 31) |
-                ((self.scans["quality_indicators"] << 4) >> 31) |
-                ((self.scans["quality_indicators"] << 5) >> 31))
-        return mask.astype(bool)
-
-    def get_qual_flags(self):
-        """Read quality flags."""
-        number_of_scans = self.scans["telemetry"].shape[0]
-        qual_flags = np.zeros((int(number_of_scans), 7))
-        qual_flags[:, 0] = self.scans["scan_line_number"]
-        qual_flags[:, 1] = (self.scans["quality_indicators"] >> 31)
-        qual_flags[:, 2] = ((self.scans["quality_indicators"] << 4) >> 31)
-        qual_flags[:, 3] = ((self.scans["quality_indicators"] << 5) >> 31)
-        qual_flags[:, 4] = ((self.scans["quality_indicators"] << 13) >> 31)
-        qual_flags[:, 5] = ((self.scans["quality_indicators"] << 14) >> 31)
-        qual_flags[:, 6] = ((self.scans["quality_indicators"] << 15) >> 31)
-
-        return qual_flags
-
     def postproc(self, channels):
         """No POD specific postprocessing to be done."""
         pass
@@ -465,31 +539,26 @@ class PODReader(Reader):
         return get_tsm_idx(channels[:, :, 0], channels[:, :, 1],
                            channels[:, :, 3], channels[:, :, 4])
 
+    def _get_calibrated_channels_uniform_shape(self):
+        """Prepare the channels as input for gac_io.save_gac"""
+        _channels = self.get_calibrated_channels()
+        # prepare input
+        # maybe there is a better (less memory requiring) method
+        shape = _channels.shape[:-1] + (6,)
+        # empty_like does not know the kwarg shape in older versions
+        channels = np.empty(shape, dtype=_channels.dtype)
+        channels[:, :, 0] = _channels[:, :, 0]
+        channels[:, :, 1] = _channels[:, :, 1]
+        channels[:, :, 2] = np.nan
+        channels[:, :, 3] = _channels[:, :, 2]
+        channels[:, :, 4] = _channels[:, :, 3]
+        channels[:, :, 5] = _channels[:, :, 4]
+        return channels
+
 
 def main_pod(reader_cls, filename, start_line, end_line):
     """Generate a l1c file."""
-    from pygac import gac_io
     tic = datetime.datetime.now()
-    reader = reader_cls()
-    reader.read(filename)
-    reader.get_lonlat()
-    channels = reader.get_calibrated_channels()
-    sat_azi, sat_zen, sun_azi, sun_zen, rel_azi = reader.get_angles()
-
-    qual_flags = reader.get_qual_flags()
-    if (np.all(reader.mask)):
-        print("ERROR: All data is masked out. Stop processing")
-        raise ValueError("All data is masked out.")
-    gac_io.save_gac(reader.spacecraft_name,
-                    reader.utcs,
-                    reader.lats, reader.lons,
-                    channels[:, :, 0], channels[:, :, 1],
-                    np.full_like(channels[:, :, 0], np.nan),
-                    channels[:, :, 2],
-                    channels[:, :, 3],
-                    channels[:, :, 4],
-                    sun_zen, sat_zen, sun_azi, sat_azi, rel_azi,
-                    qual_flags, start_line, end_line,
-                    reader.filename,
-                    reader.meta_data)
+    reader = reader_cls.fromfile(filename)
+    reader.save(int(start_line), int(end_line))
     LOG.info("pygac took: %s", str(datetime.datetime.now() - tic))
