@@ -1,5 +1,4 @@
 #!/usr/bin/env python
-# -*- coding: utf-8 -*-
 
 # Copyright (c) 2014, 2019 Pygac Developers
 
@@ -26,26 +25,22 @@
 Can't be used as is, has to be subclassed to add specific read functions.
 """
 from abc import ABCMeta, abstractmethod, abstractproperty
+import datetime
 import logging
 import numpy as np
 import os
+import re
 import six
 import types
-import gzip
-from contextlib import contextmanager
-
-from pygac import (CONFIG_FILE, centered_modulus,
-                   calculate_sun_earth_distance_correction,
-                   get_absolute_azimuth_angle_diff)
-try:
-    import ConfigParser
-except ImportError:
-    import configparser as ConfigParser
+import warnings
 
+from pygac.utils import (centered_modulus,
+                         calculate_sun_earth_distance_correction,
+                         get_absolute_azimuth_angle_diff)
 from pyorbital.orbital import Orbital
 from pyorbital import astronomy
-import datetime
-from pygac.calibration import calibrate_solar, calibrate_thermal
+from pygac.calibration import Calibrator, calibrate_solar, calibrate_thermal
+from pygac import gac_io
 
 LOG = logging.getLogger(__name__)
 
@@ -78,11 +73,21 @@ rpy_coeffs = {
                }}
 
 
+class ReaderError(ValueError):
+    """Raised in Reader.read if the given file does not correspond to it"""
+    pass
+
+
 class Reader(six.with_metaclass(ABCMeta)):
     """Reader for Gac and Lac, POD and KLM data."""
 
+    # data set header format, see _validate_header for more details
+    data_set_pattern = re.compile(
+        r'\w{3}\.\w{4}\.\w{2}.D\d{5}\.S\d{4}\.E\d{4}\.B\d{7}\.\w{2}')
+
     def __init__(self, interpolate_coords=True, adjust_clock_drift=True,
-                 tle_dir=None, tle_name=None, tle_thresh=7, creation_site=None):
+                 tle_dir=None, tle_name=None, tle_thresh=7, creation_site=None,
+                 custom_calibration=None, calibration_file=None):
         """Init the reader.
 
         Args:
@@ -95,6 +100,9 @@ class Reader(six.with_metaclass(ABCMeta)):
             tle_thresh: Maximum number of days between observation and nearest
                 TLE
             creation_site: The three-letter identifier of the creation site (eg 'NSS')
+            custom_calibration: dictionary with a subset of user defined satellite specific
+                                calibration coefficients
+            calibration_file: path to json file containing default calibrations
 
         """
         self.meta_data = {}
@@ -104,6 +112,8 @@ class Reader(six.with_metaclass(ABCMeta)):
         self.tle_name = tle_name
         self.tle_thresh = tle_thresh
         self.creation_site = (creation_site or 'NSS').encode('utf-8')
+        self.custom_calibration = custom_calibration
+        self.calibration_file = calibration_file
         self.head = None
         self.scans = None
         self.spacecraft_name = None
@@ -116,38 +126,219 @@ class Reader(six.with_metaclass(ABCMeta)):
         self.filename = None
         self._mask = None
 
+    @property
+    def filename(self):
+        """Get the property 'filename'."""
+        return self._filename
+
+    @filename.setter
+    def filename(self, filepath):
+        """Set the property 'filename'."""
+        if filepath is None:
+            self._filename = None
+        else:
+            match = self.data_set_pattern.search(filepath)
+            if match:
+                self._filename = match.group()
+            else:
+                self._filename = os.path.basename(filepath)
+
+    @property
+    def calibration(self):
+        """Get the property 'calibration'."""
+        calibration = Calibrator(
+            self.spacecraft_name,
+            custom_coeffs=self.custom_calibration,
+            coeffs_file=self.calibration_file
+        )
+        return calibration
+
     @abstractmethod
-    def read(self, filename):
+    def read(self, filename, fileobj=None):
         """Read the GAC/LAC data.
 
         Args:
-            filename (str): Specifies the GAC/LAC file to be read.
+            filename (str): Path to GAC/LAC file
+            fileobj: An open file object to read from. (optional)
         """
         raise NotImplementedError
 
-    @contextmanager
-    def _open(self, filename):
-        """Open the GAC/LAC data file and yield the filehandle.
+    @classmethod
+    @abstractmethod
+    def read_header(cls, filename, fileobj=None):
+        """Read the file header.
 
         Args:
-            filename (str): Path to GAC/LAC file to open
+            filename (str): Path to GAC/LAC file
+            fileobj: An open file object to read from. (optional)
+
+        Returns:
+            archive_header (struct): archive header
+            header (struct): file header
 
         Note:
-            This method should be called inside the Reader.read implementation
+            This is a classmethod to avoid throwaway instances while
+            checking if the reader corresponds to the input file.
         """
-        basename = os.path.basename(filename)
-        root, ext = os.path.splitext(basename)
-        if ext == ".gz":
-            self.filename = root
-            file_object = gzip.open(filename, mode='rb')
-        else:
-            self.filename = basename
-            file_object = open(filename, mode='rb')
-        LOG.info('Reading %s', self.filename)
+        raise NotImplementedError
+
+    @classmethod
+    def _correct_data_set_name(cls, header, filename):
+        """Replace invalid data_set_name from header with filename
+
+        Args:
+            header (struct): file header
+            filename (str): path to file
+        """
+        data_set_name = header['data_set_name'].decode(errors='ignore')
+        if not cls.data_set_pattern.match(data_set_name):
+            LOG.debug('The data_set_name in header %s does not match.'
+                      ' Use filename instead.' % header['data_set_name'])
+            match = cls.data_set_pattern.search(filename)
+            if match:
+                data_set_name = match.group()
+                LOG.debug("Set data_set_name, to filename %s"
+                          % data_set_name)
+                header['data_set_name'] = data_set_name.encode()
+            else:
+                LOG.debug("header['data_set_name']=%s; filename='%s'"
+                          % (header['data_set_name'], filename))
+                raise ReaderError('Cannot determine data_set_name!')
+        return header
+
+    @classmethod
+    def _validate_header(cls, header):
+        """Check if the header belongs to this reader
+
+        Note:
+            according to https://www1.ncdc.noaa.gov/pub/data/satellite/
+            publications/podguides/TIROS-N%20thru%20N-14/pdf/NCDCPOD2.pdf
+            and https://www1.ncdc.noaa.gov/pub/data/satellite/
+            publications/podguides/N-15%20thru%20N-19/pdf/
+            2.5%20Section%208.0%20NOAA%20Level%201B%20Database.pdf
+            the data set name splits into
+            PROCESSING-CENTER.DATA-TYPE.SPACECRAFT-UNIQUE-ID.
+            YEAR-DAY.START-TIME.STOP-TIME.PROCESSING-BLOCK-ID.SOURCE
+            This should be sufficient information to determine the
+            reader.
+            Global Area Coverage (GAC):
+                DATA-TYPE = GHRR
+            Local Area Coverage (LAC):
+                DATA-TYPE = LHRR
+            Polar Orbiter Data (POD):
+                SPACECRAFT-UNIQUE-ID in [TN, NA, NB, NC, ND, NE, NF,
+                                         NG, NH, NI, NJ]
+            NOAA-K, -L, -M system, but also newer satellites (KLM):
+                SPACECRAFT-UNIQUE-ID in [NK, NL, NM, NN, NP, M2, M1]
+        """
+        # This method does not need to be implemented in all subclasses.
+        # It is intended for cooperative multiple inheritance, i.e.
+        # each child class which implements this method, should call the
+        # super method to enter into the method resolution order.
+        # See https://docs.python.org/3/library/functions.html#super
+        # second use case "diamond diagrams".
+        # Check if the data set name matches the pattern
+        LOG.debug("validate header")
+        data_set_name = header['data_set_name'].decode(errors='ignore')
+        if not cls.data_set_pattern.match(data_set_name):
+            raise ReaderError('Data set name %s does not match!'
+                              % header['data_set_name'])
+
+    def _read_scanlines(self, buffer, count):
+        """Read the scanlines from the given buffer
+
+        Args:
+            buffer (bytes, bytearray): buffer to read from
+            count (int): number of expected scanlines
+        """
+        # Calculate the actual number of complete scanlines. The integer divisoin
+        # may strip a potentially incomplete line at the end of the file.
+        line_count = len(buffer) // self.scanline_type.itemsize
+        if line_count != count:
+            LOG.warning(
+                "Expected %d scan lines, but found %d!"
+                % (count, line_count))
+            warnings.warn("Unexpected number of scanlines!",
+                          category=RuntimeWarning)
+        self.scans = np.frombuffer(
+            buffer, dtype=self.scanline_type, count=line_count)
+
+    @classmethod
+    def can_read(cls, filename, fileobj=None):
+        """Read the GAC/LAC data.
+
+        Args:
+            filename (str): Path to GAC/LAC file
+            fileobj: An open file object to read from. (optional)
+
+        Retruns:
+            result (bool): True if the reader can read the input
+        """
+        if fileobj:
+            pos = fileobj.tell()
         try:
-            yield file_object
+            archive_header, header = cls.read_header(
+                filename, fileobj=fileobj)
+            result = True
+        except (ReaderError, ValueError) as exception:
+            LOG.debug("%s failed to read the file! %s"
+                      % (cls.__name__, repr(exception)))
+            result = False
         finally:
-            file_object.close()
+            if fileobj:
+                fileobj.seek(pos)
+        return result
+
+    @classmethod
+    def fromfile(cls, filename, fileobj=None):
+        """Create Reader from file (alternative constructor)
+
+        Args:
+            filename (str): Path to GAC/LAC file
+
+        Kwargs:
+            fileobj (file object): Open file object to read from
+
+        Note:
+            The fileobj is useful when dealing with tar archives,
+            where the filename is given by the tarinfo object, but
+            the extracted file object's property 'name' is set to
+            the filename of the archive.
+        """
+        instance = cls()
+        instance.read(filename, fileobj=fileobj)
+        return instance
+
+    def _get_calibrated_channels_uniform_shape(self):
+        """Prepare the channels as input for gac_io.save_gac"""
+        channels = self.get_calibrated_channels()
+        assert channels.shape[-1] == 6
+        return channels
+
+    def save(self, start_line, end_line, output_file_prefix="PyGAC", output_dir="./",
+             avhrr_dir=None, qual_dir=None, sunsatangles_dir=None):
+        """Convert the Reader instance content into hdf5 files"""
+        avhrr_dir = avhrr_dir or output_dir
+        qual_dir = qual_dir or output_dir
+        sunsatangles_dir = sunsatangles_dir or output_dir
+        self.get_lonlat()
+        channels = self._get_calibrated_channels_uniform_shape()
+        sat_azi, sat_zen, sun_azi, sun_zen, rel_azi = self.get_angles()
+        qual_flags = self.get_qual_flags()
+        if (np.all(self.mask)):
+            print("ERROR: All data is masked out. Stop processing")
+            raise ValueError("All data is masked out.")
+        gac_io.save_gac(
+            self.spacecraft_name,
+            self.utcs, self.lats, self.lons,
+            channels[:, :, 0], channels[:, :, 1],
+            channels[:, :, 2], channels[:, :, 3],
+            channels[:, :, 4], channels[:, :, 5],
+            sun_zen, sat_zen, sun_azi, sat_azi, rel_azi,
+            qual_flags, start_line, end_line,
+            self.filename, self.meta_data,
+            output_file_prefix, avhrr_dir, qual_dir, sunsatangles_dir
+        )
 
     @abstractmethod
     def get_header_timestamp(self):
@@ -360,6 +551,9 @@ class Reader(six.with_metaclass(ABCMeta)):
             self.meta_data['midnight_scanline'] = self.get_midnight_scanline()
         if 'missing_scanlines' not in self.meta_data:
             self.meta_data['missing_scanlines'] = self.get_miss_lines()
+        if 'gac_header' not in self.meta_data:
+            self.meta_data['gac_header'] = self.head.copy()
+        self.meta_data['calib_coeffs_version'] = self.calibration.version
 
     def get_calibrated_channels(self):
         """Calibrate and return the channels."""
@@ -371,15 +565,18 @@ class Reader(six.with_metaclass(ABCMeta)):
         jday = delta.days + 1
 
         corr = self.meta_data['sun_earth_distance_correction_factor']
+        calibration_coeffs = self.calibration
 
         # how many reflective channels are there ?
         tot_ref = channels.shape[2] - 3
 
-        channels[:, :, 0:tot_ref] = calibrate_solar(channels[:, :, 0:tot_ref],
-                                                    np.arange(tot_ref),
-                                                    year, jday,
-                                                    self.spacecraft_name,
-                                                    corr)
+        channels[:, :, 0:tot_ref] = calibrate_solar(
+            channels[:, :, 0:tot_ref],
+            np.arange(tot_ref),
+            year, jday,
+            calibration_coeffs,
+            corr
+        )
         prt, ict, space = self.get_telemetry()
         for chan in [3, 4, 5]:
             channels[:, :, chan - 6] = calibrate_thermal(
@@ -389,7 +586,8 @@ class Reader(six.with_metaclass(ABCMeta)):
                 space[:, chan - 3],
                 self.scans["scan_line_number"],
                 chan,
-                self.spacecraft_name)
+                calibration_coeffs
+            )
 
         # Mask out corrupt values
         channels[self.mask] = np.nan
@@ -444,10 +642,34 @@ class Reader(six.with_metaclass(ABCMeta)):
             self._mask = self._get_corrupt_mask()
         return self._mask
 
-    @abstractmethod
-    def _get_corrupt_mask(self):
-        """KLM/POD specific readout of corrupt scanline mask."""
-        raise NotImplementedError
+    def _get_corrupt_mask(self, flags=None):
+        """Readout of corrupt scanline mask.
+
+        Args:
+            flags (QFlag.flag): An "ORed" bitmask that defines corrupt values.
+                                Defauts to (QFlag.FATAL_FLAG | QFlag.CALIBRATION
+                                | QFlag.NO_EARTH_LOCATION)
+
+        Note:
+            The Quality flags mapping (QFlag) is KLM/POD specific.
+        """
+        QFlag = self.QFlag
+        if flags is None:
+            flags = QFlag.FATAL_FLAG | QFlag.CALIBRATION | QFlag.NO_EARTH_LOCATION
+        return (self.scans[self._quality_indicators_key] & int(flags)).astype(bool)
+
+    def get_qual_flags(self):
+        """Read quality flags."""
+        number_of_scans = self.scans["telemetry"].shape[0]
+        qual_flags = np.zeros((int(number_of_scans), 7))
+        qual_flags[:, 0] = self.scans["scan_line_number"]
+        qual_flags[:, 1] = self._get_corrupt_mask(flags=self.QFlag.FATAL_FLAG)
+        qual_flags[:, 2] = self._get_corrupt_mask(flags=self.QFlag.CALIBRATION)
+        qual_flags[:, 3] = self._get_corrupt_mask(flags=self.QFlag.NO_EARTH_LOCATION)
+        qual_flags[:, 4] = self._get_corrupt_mask(flags=self.QFlag.CH_3_CONTAMINATION)
+        qual_flags[:, 5] = self._get_corrupt_mask(flags=self.QFlag.CH_4_CONTAMINATION)
+        qual_flags[:, 6] = self._get_corrupt_mask(flags=self.QFlag.CH_5_CONTAMINATION)
+        return qual_flags
 
     @abstractmethod
     def postproc(self, channels):
@@ -483,24 +705,10 @@ class Reader(six.with_metaclass(ABCMeta)):
     def get_tle_file(self):
         """Find TLE file for the current satellite."""
         tle_dir, tle_name = self.tle_dir, self.tle_name
-
-        # If user didn't specify TLE dir/name, try config file
-        if tle_dir is None or tle_name is None:
-            conf = ConfigParser.ConfigParser()
-            try:
-                conf.read(CONFIG_FILE)
-            except ConfigParser.NoSectionError:
-                LOG.exception('Failed reading configuration file: %s',
-                              str(CONFIG_FILE))
-                raise
-
-            options = {}
-            for option, value in conf.items('tle', raw=True):
-                options[option] = value
-
-            tle_dir = options['tledir']
-            tle_name = options['tlename']
-
+        if tle_dir is None:
+            raise RuntimeError("TLE directory not specified!")
+        if tle_name is None:
+            raise RuntimeError("TLE name not specified!")
         values = {"satname": self.spacecraft_name, }
         tle_filename = os.path.join(tle_dir, tle_name % values)
         LOG.info('TLE filename = ' + str(tle_filename))
