@@ -6,6 +6,7 @@
 # Author(s):
 
 #   Martin Raspaud <martin.raspaud@smhi.se>
+#   Carlos Horn <carlos.horn@external.eumetsat.int>
 
 # This program is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -25,7 +26,8 @@
 
 from pygac.tests import (test_calibrate_pod, test_slerp, test_calibrate_klm,
                          test_pod, test_tsm, test_reader, test_io,
-                         test_angles, test_init)
+                         test_angles, test_init, test_utils,
+                         test_calibration_coefficients)
 import unittest
 
 
@@ -35,7 +37,7 @@ def suite():
     mysuite = unittest.TestSuite()
     tests = (test_slerp, test_calibrate_klm, test_calibrate_pod,
              test_pod, test_tsm, test_reader, test_io, test_angles,
-             test_init)
+             test_init, test_utils, test_calibration_coefficients)
     for test in tests:
         mysuite.addTests(test.suite())
 
