@@ -6,6 +6,7 @@
 # Author(s):
 
 #   Nina Hakansson <nina.hakansson@smhi.se>
+#   Carlos Horn <carlos.horn@external.eumetsat.int>
 
 # This program is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -26,16 +27,10 @@ import unittest
 
 import numpy as np
 
-from pygac import calculate_sun_earth_distance_correction
-
 
 class TestInit(unittest.TestCase):
     """Test function for the angle calculation."""
-
-    def test_calculate_sun_earth_distance_correction(self):
-        """Test function for the sun distance corretction."""
-        corr = calculate_sun_earth_distance_correction(3)
-        np.testing.assert_almost_equal(corr, 0.96660494, decimal=7)
+    pass
 
 
 def suite():
